==> ai_parrot-0.6.0-cp310-cp310-manylinux_2_17_x86_64.manylinux2014_x86_64.whl: parrot/tools/gvoice.py <==
import asyncio
from pathlib import Path
from typing import Optional
import os
import re
import html
from xml.sax.saxutils import escape
from datetime import datetime
import aiofiles
# Use v1 for wider feature set including SSML
from google.cloud import texttospeech_v1 as texttospeech
from google.oauth2 import service_account
from langchain.tools import BaseTool
from navconfig import BASE_DIR
from parrot.conf import GOOGLE_TTS_SERVICE


class GoogleVoiceTool(BaseTool):
    """Generate a podcast-style audio file from markdown text using Google Cloud Text-to-Speech."""
    name: str = "podcast_generator_tool"
    description: str = (
        "Generates a podcast-style audio file from Markdown text using Google Cloud Text-to-Speech."
        " This tool is useful for creating audio content from text, such as articles or blog posts."
        " Provide the text in Text or markdown format, provide the text *as-is* without enclosing on backticks or quotes."
        " It can also be used to convert text to speech for accessibility purposes."
        " The audio file will be saved in the specified output directory and returned as a dictionary with a *file_path* key."
    )
    voice_model: str = "en-US-Neural2-F"  # "en-US-Studio-O"
    voice_gender: str = "FEMALE"
    language_code: str = "en-US"
    output_format: str = "OGG_OPUS"  # OGG format is more podcast-friendly
    _key_service: Optional[str]

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._key_service = GOOGLE_TTS_SERVICE
        if self.voice_gender == 'FEMALE':
            self.voice_model = "en-US-Neural2-F"
        elif self.voice_gender == 'MALE':
            self.voice_model = "en-US-Neural2-M"
        else:
            self.voice_model = "en-US-Neural2-G"

    def is_markdown(self, text: str) -> bool:
        """
        Checks if a text string is likely Markdown by looking for structural and common syntax.
        """
        if not text.strip():
            return False  # Empty text is not Markdown

        # Check for structural elements that often indicate Markdown
        if re.search(r"^#+\s", text, re.MULTILINE):  # Headings at the start of lines
            return True
        if re.search(r"^\s*([-*]\s|\d+\.\s)", text, re.MULTILINE):  # Lists at the start of lines
            return True
        if re.search(r"```", text):  # Code blocks
            return True
        if re.search(r"\[.+?\]\(.+?\)", text):  # Links
            return True
        if re.search(r"> .+", text, re.MULTILINE):  # Blockquotes
            return True
        if re.search(r"^-{3,}$|^\*{3,}$|^_{3,}$", text, re.MULTILINE):  # Horizontal rules
            return True
        if re.search(r"\*\*\w+\*\*", text) or re.search(r"__\w+__", text):  # Bold
            return True
        if re.search(r"\*\w+\*", text) or re.search(r"_\w+_", text):  # Italics
            return True
        if re.search(r"`[^`]+`", text):  # Inline code
            return True
        if re.search(r"\n\n", text.strip()): # Paragraph breaks (more than one newline)
            return True

        # If none of the stronger structural indicators are found, do a more general check
        if re.search(r"^[#*_>-\d`\[]", text.strip()[0]): # Check if the first non-space char is a common Markdown marker
            return True

        return False


    def text_to_ssml(self, text: str) -> str:
        """Converts plain text to SSML."""
        ssml = f"<speak><p>{escape(text)}</p></speak>"
        return ssml

    def markdown_to_ssml(self, markdown_text: str) -> str:
        """Converts Markdown text to SSML, handling code blocks and ellipses."""

        ssml = "<speak>"
        lines = markdown_text.split('\n')
        in_code_block = False

        for line in lines:
            line = line.strip()

            if line.startswith("```"):
                in_code_block = not in_code_block
                if in_code_block:
                    ssml += '<prosody rate="x-slow"><p><code>'
                else:
                    ssml += '</code></p></prosody>'
                continue

            if in_code_block:
                ssml += escape(line) + '<break time="100ms"/>'  # Add slight pauses within code
                continue

            if line == "...":
                ssml += '<break time="500ms"/>'  # Keep the pause for ellipses
                continue

            # Handle Markdown headings
            heading_match = re.match(r"^(#+)\s+(.*)", line)
            if heading_match:
                heading_level = len(heading_match.group(1))  # Number of '#'
                heading_text = heading_match.group(2).strip()
                ssml += f'<p><emphasis level="strong">{escape(heading_text)}</emphasis></p>'
                continue

            if line:
                ssml += f'<p>{escape(line)}</p>'

        ssml += "</speak>"
        return ssml

    async def _generate_podcast(self, query: str) -> dict:
        try:
            print("1. Converting Markdown to SSML...")
            if self.is_markdown(query):
                ssml_text = self.markdown_to_ssml(query)
            else:
                ssml_text = self.text_to_ssml(query)
            print(f"Generated SSML:\n{ssml_text}\n") # Uncomment for debugging
            print(
                f"2. Initializing Text-to-Speech client (Voice: {self.voice_model})..."
            )
            if not os.path.exists(self._key_service):
                raise FileNotFoundError(
                    f"Service account file not found: {self._key_service}"
                )
            credentials = service_account.Credentials.from_service_account_file(
                self._key_service
            )
            # Initialize the Text-to-Speech client with the service account credentials
            # Use the v1 API for wider feature set including SSML
            client = texttospeech.TextToSpeechClient(credentials=credentials)
            synthesis_input = texttospeech.SynthesisInput(ssml=ssml_text)
            # Select the voice parameters
            voice = texttospeech.VoiceSelectionParams(
                language_code=self.language_code,
                name=self.voice_model
            )
            # Select the audio format (OGG with OPUS codec)
            timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
            # Generate a unique filename based on the current timestamp
            output_filename = f"podcast_{timestamp}.ogg"  # Default output filename
            # default to OGG
            encoding = texttospeech.AudioEncoding.OGG_OPUS
            if self.output_format == "OGG_OPUS":
                encoding = texttospeech.AudioEncoding.OGG_OPUS
                output_filename = f"podcast_{timestamp}.ogg"
            elif self.output_format == "MP3":
                encoding = texttospeech.AudioEncoding.MP3
                output_filename = f"podcast_{timestamp}.mp3"
            elif self.output_format == "LINEAR16":
                encoding = texttospeech.AudioEncoding.LINEAR16
                output_filename = f"podcast_{timestamp}.wav"
            elif self.output_format == "WEBM_OPUS":
                encoding = texttospeech.AudioEncoding.WEBM_OPUS
                output_filename = f"podcast_{timestamp}.webm"
            elif self.output_format == "FLAC":
                encoding = texttospeech.AudioEncoding.FLAC
                output_filename = f"podcast_{timestamp}.flac"
            elif self.output_format == "OGG_VORBIS":
                encoding = texttospeech.AudioEncoding.OGG_VORBIS
                output_filename = f"podcast_{timestamp}.ogg"

            audio_config = texttospeech.AudioConfig(
                audio_encoding=encoding,
                speaking_rate=1.0,
                pitch=0.0
            )
            print("3. Synthesizing speech...")
            response = client.synthesize_speech(
                input=synthesis_input,
                voice=voice,
                audio_config=audio_config
            )
            print("4. Speech synthesized successfully.")
            # Get the absolute path for the output file
            output_dir: Path = BASE_DIR.joinpath('static', 'audio', 'podcasts')
            if output_dir.exists() is False:
                # Create the directory if it doesn't exist
                output_dir.mkdir(parents=True, exist_ok=True)
            output_filepath = output_dir.joinpath(output_filename)
            print(f"5. Saving audio content to: {output_filepath}")
            async with aiofiles.open(output_filepath, 'wb') as audio_file:
                await audio_file.write(response.audio_content)
            print("6. Audio content saved successfully.")
            return {
                "file_path": output_filepath,
                "output_format": self.output_format,
                "language_code": self.language_code,
                "voice_model": self.voice_model,
                "timestamp": timestamp,
                "filename": output_filename
            }
        except Exception as e:
            print(f"An error occurred: {e}")
            return {"error": str(e)}

    async def _arun(self, query: str) -> dict:
        """
        Generates a podcast-style audio file from Markdown text using
        Google Cloud Text-to-Speech.

        Args:
            markdown_summary: The input text in Markdown format.
            output_filename: The desired name for the output audio file (e.g., "my_podcast.ogg").
            language_code: The language code (e.g., "en-US", "es-ES").
            voice_name: The specific voice model name. Find names here:
                        https://cloud.google.com/text-to-speech/docs/voices

        Returns:
            A dictionary containing the absolute path to the saved audio file
            under the key "file_path", or an error message under "error".
        """
        return await self._generate_podcast(query)

    def _run(self, query: str) -> dict:
        """
        Synchronous method to generate a podcast-style audio file from Markdown text.
        This method is not recommended for production use due to blocking I/O.
        """
        loop = asyncio.get_event_loop()
        if loop.is_running():
            # If the event loop is already running, use run_until_complete
            return loop.run_until_complete(self._generate_podcast(query))
        else:
            # If not, use run
            return loop.run(self._generate_podcast(query))

==> ai_parrot-0.6.2-cp310-cp310-manylinux_2_17_x86_64.manylinux2014_x86_64.whl: parrot/tools/gvoice.py <==
import asyncio
from pathlib import Path
from typing import Optional
import os
import re
import html
from xml.sax.saxutils import escape
from datetime import datetime
import aiofiles
# Use v1 for wider feature set including SSML
from google.cloud import texttospeech_v1 as texttospeech
from google.oauth2 import service_account
from langchain.tools import BaseTool
from navconfig import BASE_DIR
from parrot.conf import GOOGLE_TTS_SERVICE


class GoogleVoiceTool(BaseTool):
    """Generate a podcast-style audio file from Text using Google Cloud Text-to-Speech."""
    name: str = "podcast_generator_tool"
    description: str = (
        "Generates a podcast-style audio file from a given text script using Google Cloud Text-to-Speech."
        " This tool is useful for creating audio content from text, such as articles or blog posts."
        " It can also be used to convert text to speech for accessibility purposes."
        " Provide the text *as-is* without enclosing on backticks or backquotes."
        " The audio file will be saved in the specified output directory and returned as a dictionary with a *file_path* key."
    )
    voice_model: str = "en-US-Neural2-F"  # "en-US-Studio-O"
    voice_gender: str = "FEMALE"
    language_code: str = "en-US"
    output_format: str = "OGG_OPUS"  # OGG format is more podcast-friendly
    _key_service: Optional[str]

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._key_service = GOOGLE_TTS_SERVICE
        if self.voice_gender == 'FEMALE':
            self.voice_model = "en-US-Neural2-F"
        elif self.voice_gender == 'MALE':
            self.voice_model = "en-US-Neural2-M"
        else:
            self.voice_model = "en-US-Neural2-G"

    def is_markdown(self, text: str) -> bool:
        """
        Checks if a text string is likely Markdown by looking for structural and common syntax.
        """
        if not text.strip():
            return False  # Empty text is not Markdown

        # Check for structural elements that often indicate Markdown
        if re.search(r"^#+\s", text, re.MULTILINE):  # Headings at the start of lines
            return True
        if re.search(r"^\s*([-*]\s|\d+\.\s)", text, re.MULTILINE):  # Lists at the start of lines
            return True
        if re.search(r"```", text):  # Code blocks
            return True
        if re.search(r"\[.+?\]\(.+?\)", text):  # Links
            return True
        if re.search(r"> .+", text, re.MULTILINE):  # Blockquotes
            return True
        if re.search(r"^-{3,}$|^\*{3,}$|^_{3,}$", text, re.MULTILINE):  # Horizontal rules
            return True
        if re.search(r"\*\*\w+\*\*", text) or re.search(r"__\w+__", text):  # Bold
            return True
        if re.search(r"\*\w+\*", text) or re.search(r"_\w+_", text):  # Italics
            return True
        if re.search(r"`[^`]+`", text):  # Inline code
            return True
        if re.search(r"\n\n", text.strip()): # Paragraph breaks (more than one newline)
            return True

        # If none of the stronger structural indicators are found, do a more general check
        if re.search(r"^[#*_>-\d`\[]", text.strip()[0]): # Check if the first non-space char is a common Markdown marker
            return True

        return False


    def text_to_ssml(self, text: str) -> str:
        """Converts plain text to SSML."""
        ssml = f"<speak><p>{escape(text)}</p></speak>"
        return ssml

    def markdown_to_ssml(self, markdown_text: str) -> str:
        """Converts Markdown text to SSML, handling code blocks and ellipses."""

        if markdown_text.startswith("```text"):
            markdown_text = markdown_text[len("```text"):].strip()

        ssml = "<speak>"
        lines = markdown_text.split('\n')
        in_code_block = False

        for line in lines:
            line = line.strip()

            if line.startswith("```"):
                in_code_block = not in_code_block
                if in_code_block:
                    ssml += '<prosody rate="x-slow"><p><code>'
                else:
                    ssml += '</code></p></prosody>'
                continue

            if in_code_block:
                ssml += escape(line) + '<break time="100ms"/>'  # Add slight pauses within code
                continue

            if line == "...":
                ssml += '<break time="500ms"/>'  # Keep the pause for ellipses
                continue

            # Handle Markdown headings
            heading_match = re.match(r"^(#+)\s+(.*)", line)
            if heading_match:
                heading_level = len(heading_match.group(1))  # Number of '#'
                heading_text = heading_match.group(2).strip()
                ssml += f'<p><emphasis level="strong">{escape(heading_text)}</emphasis></p>'
                continue

            if line:
                ssml += f'<p>{escape(line)}</p>'

        ssml += "</speak>"
        return ssml

    async def _generate_podcast(self, query: str) -> dict:
        try:
            print("1. Converting Markdown to SSML...")
            if self.is_markdown(query):
                ssml_text = self.markdown_to_ssml(query)
            else:
                ssml_text = self.text_to_ssml(query)
            print(f"Generated SSML:\n{ssml_text}\n") # Uncomment for debugging
            print(
                f"2. Initializing Text-to-Speech client (Voice: {self.voice_model})..."
            )
            if not os.path.exists(self._key_service):
                raise FileNotFoundError(
                    f"Service account file not found: {self._key_service}"
                )
            credentials = service_account.Credentials.from_service_account_file(
                self._key_service
            )
            # Initialize the Text-to-Speech client with the service account credentials
            # Use the v1 API for wider feature set including SSML
            client = texttospeech.TextToSpeechClient(credentials=credentials)
            synthesis_input = texttospeech.SynthesisInput(ssml=ssml_text)
            # Select the voice parameters
            voice = texttospeech.VoiceSelectionParams(
                language_code=self.language_code,
                name=self.voice_model
            )
            # Select the audio format (OGG with OPUS codec)
            timestamp = datetime.now().strftime("%Y%m%d_%H%M%S")
            # Generate a unique filename based on the current timestamp
            output_filename = f"podcast_{timestamp}.ogg"  # Default output filename
            # default to OGG
            encoding = texttospeech.AudioEncoding.OGG_OPUS
            if self.output_format == "OGG_OPUS":
                encoding = texttospeech.AudioEncoding.OGG_OPUS
                output_filename = f"podcast_{timestamp}.ogg"
            elif self.output_format == "MP3":
                encoding = texttospeech.AudioEncoding.MP3
                output_filename = f"podcast_{timestamp}.mp3"
            elif self.output_format == "LINEAR16":
                encoding = texttospeech.AudioEncoding.LINEAR16
                output_filename = f"podcast_{timestamp}.wav"
            elif self.output_format == "WEBM_OPUS":
                encoding = texttospeech.AudioEncoding.WEBM_OPUS
                output_filename = f"podcast_{timestamp}.webm"
            elif self.output_format == "FLAC":
                encoding = texttospeech.AudioEncoding.FLAC
                output_filename = f"podcast_{timestamp}.flac"
            elif self.output_format == "OGG_VORBIS":
                encoding = texttospeech.AudioEncoding.OGG_VORBIS
                output_filename = f"podcast_{timestamp}.ogg"

            audio_config = texttospeech.AudioConfig(
                audio_encoding=encoding,
                speaking_rate=1.0,
                pitch=0.0
            )
            print("3. Synthesizing speech...")
            response = client.synthesize_speech(
                input=synthesis_input,
                voice=voice,
                audio_config=audio_config
            )
            print("4. Speech synthesized successfully.")
            # Get the absolute path for the output file
            output_dir: Path = BASE_DIR.joinpath('static', 'audio', 'podcasts')
            if output_dir.exists() is False:
                # Create the directory if it doesn't exist
                output_dir.mkdir(parents=True, exist_ok=True)
            output_filepath = output_dir.joinpath(output_filename)
            print(f"5. Saving audio content to: {output_filepath}")
            async with aiofiles.open(output_filepath, 'wb') as audio_file:
                await audio_file.write(response.audio_content)
            print("6. Audio content saved successfully.")
            return {
                "file_path": output_filepath,
                "output_format": self.output_format,
                "language_code": self.language_code,
                "voice_model": self.voice_model,
                "timestamp": timestamp,
                "filename": output_filename
            }
        except Exception as e:
            print(f"An error occurred: {e}")
            return {"error": str(e)}

    async def _arun(self, query: str) -> dict:
        """
        Generates a podcast-style audio file from Markdown text using
        Google Cloud Text-to-Speech.

        Args:
            markdown_summary: The input text in Markdown format.
            output_filename: The desired name for the output audio file (e.g., "my_podcast.ogg").
            language_code: The language code (e.g., "en-US", "es-ES").
            voice_name: The specific voice model name. Find names here:
                        https://cloud.google.com/text-to-speech/docs/voices

        Returns:
            A dictionary containing the absolute path to the saved audio file
            under the key "file_path", or an error message under "error".
        """
        return await self._generate_podcast(query)

    def _run(self, query: str) -> dict:
        """
        Synchronous method to generate a podcast-style audio file from Markdown text.
        This method is not recommended for production use due to blocking I/O.
        """
        loop = asyncio.get_event_loop()
        if loop.is_running():
            # If the event loop is already running, use run_until_complete
            return loop.run_until_complete(self._generate_podcast(query))
        else:
            # If not, use run
            return loop.run(self._generate_podcast(query))
